##########################
### complile quiz bank ###
##########################
# Format is:
# \begin{shinyQuizBank}{myBankName}
# \begin{shinyMCQ}{my multiple choice question text}
#   \shinyMCAR{my MC answer text}{my MC response text}
#   ...
# \end{shinyMCQ}
# ...
# \end{shinyQuizBank}
#
compileQuizBank = function(fname) {
  #if (!require("shiny")) stop("Please install the 'shiny' package and try again.")
  #if (!require("stringr")) stop("please install R package 'stringr'")
  
  ### read .tex file and remove comments
  fname = checkFile(fname, ending=".tex")
  quizText = readLines(fname)
  comments = grep("^[%]", quizText)
  if (length(comments)>0) quizText[comments] = ""
  comments = regexpr("[^\\][%]", paste0(" ", quizText))
  Sel = (comments != -1)
  if (any(Sel)) quizText[Sel] = substring(quizText[Sel], 1, comments[Sel]-1)
  
  ### find quiz bank, verify structure, and get quiz bank name and options
  # Pull off \shinyQuizBank as --> full match | bank name | bank options
  FULL_MATCH = 1; BANK_NAME = 2; BANK_OPTIONS = 3
  quizBankB = str_match(quizText, "[\\\\]begin[[:blank:]]*[{]shinyQuizBank[}][[:blank:]]*[{](.*?)[}]([{].+?[}])?")
  quizBankBL = which(!is.na(quizBankB[, FULL_MATCH]))
  if (length(quizBankBL)==0) stop("no \\begin{quizBank}{bankName} in ", fname)
  if (length(quizBankBL)>1) stop("duplicate \\begin{quizBank}{bankName}s in ", fname)
  quizBankEL = grep("[\\]end[[:blank:]]*[{]shinyQuizBank[}]", quizText)
  if (length(quizBankEL)==0) stop("no \\end{quizBank}{bankName} in ", fname)
  if (length(quizBankEL)>1) stop("duplicate \\end{quizBank}{bankName}s in ", fname)
  if (quizBankBL>quizBankEL) stop("\\end{quizBank} before \\begin{quizBank}s in ", fname)
  #
  quizBankName = quizBankB[quizBankBL, BANK_NAME]
  if (quizBankName != make.names(quizBankName)) {
    stop("invalid quiz bank name '", quizBankName, 
         "'' (must start with a letter and have no blanks or punctuation)")
  }
  quizBankOptions = quizBankB[quizBankBL, BANK_OPTIONS]
  
  ### Locate the individual questions ###
  # Quiz counter to give each radioButton control a different name (across all banks)
  questionsSoFar = 0
  #
  # discard pre/postamble
  quizText = quizText[(quizBankBL+1):(quizBankEL-1)]
  lineNumOffset = quizBankBL
  #
  # find begin and end lines for each question
  # Parse \shinyMCQ{question text}{options} -> full match | {question text} | question text | {option} | option
  FULL_MATCH = 1; QUESTION_TEXT = 3; QUESTION_OPTION = 5
  thisQuizB = str_match(quizText,
                        "[\\\\]begin[[:blank:]]*[{]shinyMCQ[}][[:blank:]]*([{](.*?)[}])[[:blank:]]*([{](.*)[}])?")
  thisQuizE = grep("[\\]end[[:blank:]]*[{]shinyMCQ[}]", quizText)
  thisQuizBL = which(!is.na(thisQuizB[, FULL_MATCH]))
  thisQuizBE = c(thisQuizBL, thisQuizE)
  if (length(thisQuizBE)==0) {
    stop("quiz bank ", quizBankName, " is empty")
  }
  if (length(thisQuizBL) != length(thisQuizE)) {
    stop('@', paste(sort(c(thisQuizE,thisQuizBL))+lineNumOffset, collapse='/'), 
         ' un-balanced begin/end shinyMCQ statements')
  }
  # record state of "fixed" option for all questions
  questionOptions = thisQuizB[thisQuizBL, QUESTION_OPTION]
  questionOptions[is.na(questionOptions)] = ""
  questionOptions = str_trim(questionOptions)
  fixed = (regexpr("f(i(x(e(d?)?)?)?)?$", questionOptions, ignore.case=TRUE) != -1)
  #
  # make matrix of begin and end lines per question
  BEL = cbind(thisQuizBL, thisQuizE)
  
  # Prevent intertwined questions
  nMC = nrow(BEL)
  if (nMC>1) {
    temp = BEL[2:nMC,1] < BEL[1:(nMC-1),2]
    if (any(temp)) {
      temp = which(temp)[1]
      stop('@', paste(BEL[temp+1,1]+lineNumOffset, 
                      BEL[temp,2]+lineNumOffset, sep='/'), 
           ' intertwined shinyMCQ statements')
    }
  }
  
  # Get all question texts for this quiz bank
  thisQuizQuestionText = thisQuizB[thisQuizBL,QUESTION_TEXT]
  
  # Warnings for misuse
  temp = (questionOptions != "" & !fixed)
  if (any(temp)) {
    warning(thisQuizB[thisQuizBL[temp][1]], " (the only option allowed is 'fixed'")
  }
  #
  temp = unlist(lapply(1:nrow(BEL), function(x) BEL[x,1]:BEL[x,2]))
  temp = setdiff(1:length(quizText), temp)
  if (length(temp)>0) {
    tempT = quizText[temp]
    tempP = intersect(grep("^[[:blank:]]*$", tempT, invert=TRUE),
                      grep("^[[:blank:]]*[%]", tempT, invert=TRUE))
    if (length(tempP) > 0) {
      stop("extraneous text found between questions for test bank '", quizBankName, "':",
           "'", tempT[tempP][1] , "'")
    }
  }
  
  # get quiz answers and responses for all questions into a list
  # (includes extensive detailed feedback on malformed questions)
  thisQuizResponses = lapply(1:nMC,
                             function(index) {
                               if (BEL[index,2] == BEL[index,1]+1) {
                                 warning('question ', index, ' of quiz bank ', '"', quizBankName,
                                         '" has no responses')
                                 return(NULL)
                               }
                               text = quizText[(BEL[index,1]+1):(BEL[index,2]-1)]
                               text = text[grep("^[[:blank:]]*$", text, invert=TRUE)]
                               if (length(text)==0) {
                                 warning('question ', index, ' of quiz bank ', '"', quizBankName,
                                         '" has no responses')
                                 return(NULL)
                               }
                               sel = grep("[\\]shinyMCAR[*]?", text)
                               if (length(sel) == 0) {
                                 warning('question ', index, ' of quiz bank ', '"', quizBankName,
                                         '" has no responses')
                                 return(NULL)
                               }
                               if (sel[1] != 1) {
                                 stop('question ', index, ' of quiz bank ', '"', quizBankName,
                                      '" does not start with \\shinyMCAR')
                               }
                               if (length(sel) == 1) {
                                 selEnd = length(text)
                               } else {
                                 selEnd = c(sel[2:length(sel)]-1, length(text))
                               }
                               return(t(sapply(1:length(sel),
                                               function(ind2) {
                                                 one = paste(text[sel[ind2]:selEnd[ind2]], collapse="")
                                                 m = str_match(one, 
                                                               "[{](.*?)[}][[:blank:]]*[{](.*?)[}][[:blank:]]*([^%]*)?")[1,]
                                                 if (is.na(m[1])) {
                                                   stop("question ", index, " of quiz bank '", quizBankName,
                                                        "' has malformed \\shinyMCAR #", ind2)
                                                 }
                                                 if (m[4]!="") {
                                                   stop("invalid text in question ", index, " of quiz bank '", quizBankName,
                                                        "' after \\shinyMCAR #", ind2, "': '", str_trim(m[4]), "'")
                                                 }
                                                 return(m[-1])
                                               })))
                             })
  assure(all(!is.null(thisQuizResponses)), 33991)
  nQuestions = length(thisQuizResponses)
  
  # store all quiz question Shiny HTML in a list named 'thisQuiz' in fooQuiz.RData (one for each quiz bank)
  thisQuiz = vector("list", nQuestions)
  #
  for (i in 1:nQuestions) {
    questionsSoFar = questionsSoFar + 1
    thisQuiz[[i]] = quizQuestion(quizBankName, thisQuizQuestionText[i], 
                                 questionChoices = thisQuizResponses[[i]][,1], 
                                 explanations = as.list(thisQuizResponses[[i]][,2]), questionsSoFar)
  }
  
  # Store all results in a binary "save" file
  attr(thisQuiz, "fixed") = fixed
  fname = paste0(quizBankName, "Quiz.RData")
  save(thisQuiz, file=fname)
  cat(fname, "successfully created.\n")
  invisible(NULL)
} # end compileQuizBank()
